@@ -1,25 +1,17 @@
 import asyncio
-from email.utils import formatdate
 import http
 import logging
-import time
-import traceback
-from urllib.parse import unquote
-from uvicorn.protocols.utils import get_local_addr, get_remote_addr, is_ssl
+import urllib
 
 import httptools
 
-
-def _get_default_headers():
-    current_time = time.time()
-    current_date = formatdate(current_time, usegmt=True).encode()
-    return b"".join([b"server: uvicorn\r\ndate: ", current_date, b"\r\n"])
+from uvicorn.protocols.utils import get_local_addr, get_remote_addr, is_ssl
 
 
 def _get_status_line(status_code):
     try:
         phrase = http.HTTPStatus(status_code).phrase.encode()
-    except ValueError:
+    except ValueError as exc:
         phrase = b""
     return b"".join([b"HTTP/1.1 ", str(status_code).encode(), b" ", phrase, b"\r\n"])
 
@@ -28,8 +20,6 @@ STATUS_LINE = {
     status_code: _get_status_line(status_code) for status_code in range(100, 600)
 }
 
-DEFAULT_HEADERS = _get_default_headers()
-
 HIGH_WATER_LIMIT = 65536
 
 
@@ -84,35 +74,29 @@ class ServiceUnavailable:
 
 
 class HttpToolsProtocol(asyncio.Protocol):
-    def __init__(
-        self,
-        app,
-        loop=None,
-        connections=None,
-        tasks=None,
-        state=None,
-        logger=None,
-        access_log=True,
-        ws_protocol_class=None,
-        root_path="",
-        limit_concurrency=None,
-        timeout_keep_alive=5,
-    ):
-        self.app = app
-        self.loop = loop or asyncio.get_event_loop()
-        self.connections = set() if connections is None else connections
-        self.tasks = set() if tasks is None else tasks
-        self.state = {"total_requests": 0} if state is None else state
-        self.logger = logger or logging.getLogger("uvicorn")
-        self.access_log = access_log and (self.logger.level <= logging.INFO)
+    def __init__(self, config, server_state):
+        if not config.loaded:
+            config.load()
+
+        self.config = config
+        self.app = config.loaded_app
+        self.loop = config.loop_instance
+        self.logger = config.logger_instance
+        self.access_log = config.access_log and (self.logger.level <= logging.INFO)
         self.parser = httptools.HttpRequestParser(self)
-        self.ws_protocol_class = ws_protocol_class
-        self.root_path = root_path
-        self.limit_concurrency = limit_concurrency
+        self.ws_protocol_class = config.ws_protocol_class
+        self.root_path = config.root_path
+        self.limit_concurrency = config.limit_concurrency
 
         # Timeouts
         self.timeout_keep_alive_task = None
-        self.timeout_keep_alive = timeout_keep_alive
+        self.timeout_keep_alive = config.timeout_keep_alive
+
+        # Global state
+        self.server_state = server_state
+        self.connections = server_state.connections
+        self.tasks = server_state.tasks
+        self.default_headers = server_state.default_headers
 
         # Per-connection state
         self.transport = None
@@ -130,11 +114,6 @@ class HttpToolsProtocol(asyncio.Protocol):
         self.cycle = None
         self.message_event = asyncio.Event()
 
-    @classmethod
-    def tick(cls):
-        global DEFAULT_HEADERS
-        DEFAULT_HEADERS = _get_default_headers()
-
     # Protocol interface
     def connection_made(self, transport):
         self.connections.add(self)
@@ -168,11 +147,11 @@ class HttpToolsProtocol(asyncio.Protocol):
 
         try:
             self.parser.feed_data(data)
-        except httptools.parser.errors.HttpParserError:
+        except httptools.parser.errors.HttpParserError as exc:
             msg = "Invalid HTTP request received."
             self.logger.warning(msg)
             self.transport.close()
-        except httptools.HttpParserUpgrade:
+        except httptools.HttpParserUpgrade as exc:
             self.handle_upgrade()
 
     def handle_upgrade(self):
@@ -181,42 +160,45 @@ class HttpToolsProtocol(asyncio.Protocol):
             if name == b"upgrade":
                 upgrade_value = value.lower()
 
-        if upgrade_value != b'websocket' or self.ws_protocol_class is None:
+        if upgrade_value != b"websocket" or self.ws_protocol_class is None:
             msg = "Unsupported upgrade request."
             self.logger.warning(msg)
-            content = [STATUS_LINE[400], DEFAULT_HEADERS]
-            content.extend([
-                b"content-type: text/plain; charset=utf-8\r\n",
-                b"content-length: " + str(len(msg)).encode('ascii') + b"\r\n",
-                b"connection: close\r\n",
-                b"\r\n",
-                msg.encode('ascii')
-            ])
+            content = [STATUS_LINE[400]]
+            for name, value in self.default_headers:
+                content.extend([name, b": ", value, b"\r\n"])
+            content.extend(
+                [
+                    b"content-type: text/plain; charset=utf-8\r\n",
+                    b"content-length: " + str(len(msg)).encode("ascii") + b"\r\n",
+                    b"connection: close\r\n",
+                    b"\r\n",
+                    msg.encode("ascii"),
+                ]
+            )
             self.transport.write(b"".join(content))
             self.transport.close()
             return
 
         self.connections.discard(self)
-        method = self.scope['method'].encode()
-        output = [method, b' ', self.url, b' HTTP/1.1\r\n']
-        for name, value in self.scope['headers']:
+        method = self.scope["method"].encode()
+        output = [method, b" ", self.url, b" HTTP/1.1\r\n"]
+        for name, value in self.scope["headers"]:
             output += [name, b": ", value, b"\r\n"]
-        output.append(b'\r\n')
+        output.append(b"\r\n")
         protocol = self.ws_protocol_class(
-            app=self.app,
-            connections=self.connections,
-            tasks=self.tasks,
-            loop=self.loop,
-            logger=self.logger
+            config=self.config, server_state=self.server_state
         )
         protocol.connection_made(self.transport)
-        protocol.data_received(b''.join(output))
+        protocol.data_received(b"".join(output))
         self.transport.set_protocol(protocol)
 
     # Parser callbacks
     def on_url(self, url):
         method = self.parser.get_method()
         parsed_url = httptools.parse_url(url)
+        path = parsed_url.path.decode("ascii")
+        if "%" in path:
+            path = urllib.parse.unquote(path)
         self.url = url
         self.expect_100_continue = False
         self.headers = []
@@ -228,7 +210,7 @@ class HttpToolsProtocol(asyncio.Protocol):
             "scheme": self.scheme,
             "method": method.decode("ascii"),
             "root_path": self.root_path,
-            "path": parsed_url.path.decode("ascii"),
+            "path": path,
             "query_string": parsed_url.query if parsed_url.query else b"",
             "headers": self.headers,
         }
@@ -264,6 +246,7 @@ class HttpToolsProtocol(asyncio.Protocol):
             flow=self.flow,
             logger=self.logger,
             access_log=self.access_log,
+            default_headers=self.default_headers,
             message_event=self.message_event,
             expect_100_continue=self.expect_100_continue,
             keep_alive=http_version != "1.0",
@@ -295,7 +278,7 @@ class HttpToolsProtocol(asyncio.Protocol):
 
     def on_response_complete(self):
         # Callback for pipelined HTTP requests to be started.
-        self.state["total_requests"] += 1
+        self.server_state.total_requests += 1
 
         if self.transport.is_closing():
             return
@@ -352,6 +335,7 @@ class RequestResponseCycle:
         flow,
         logger,
         access_log,
+        default_headers,
         message_event,
         expect_100_continue,
         keep_alive,
@@ -362,6 +346,7 @@ class RequestResponseCycle:
         self.flow = flow
         self.logger = logger
         self.access_log = access_log
+        self.default_headers = default_headers
         self.message_event = message_event
         self.on_response = on_response
 
@@ -385,10 +370,9 @@ class RequestResponseCycle:
         try:
             asgi = app(self.scope)
             result = await asgi(self.receive, self.send)
-        except:
-            msg = "Exception in ASGI application\n%s"
-            traceback_text = "".join(traceback.format_exc())
-            self.logger.error(msg, traceback_text)
+        except BaseException as exc:
+            msg = "Exception in ASGI application\n"
+            self.logger.error(msg, exc_info=exc)
             if not self.response_started:
                 await self.send_500_response()
             else:
@@ -444,20 +428,20 @@ class RequestResponseCycle:
             self.waiting_for_100_continue = False
 
             status_code = message["status"]
-            headers = message.get("headers", [])
+            headers = self.default_headers + message.get("headers", [])
 
             if self.access_log:
                 self.logger.info(
                     '%s - "%s %s HTTP/%s" %d',
                     self.scope["client"],
                     self.scope["method"],
-                    self.scope["path"],
+                    self.scope["root_path"] + self.scope["path"],
                     self.scope["http_version"],
                     status_code,
                 )
 
             # Write response status line and headers
-            content = [STATUS_LINE[status_code], DEFAULT_HEADERS]
+            content = [STATUS_LINE[status_code]]
 
             for name, value in headers:
                 name = name.lower()
@@ -471,7 +455,11 @@ class RequestResponseCycle:
                     self.keep_alive = False
                 content.extend([name, b": ", value, b"\r\n"])
 
-            if self.chunked_encoding is None:
+            if (
+                self.chunked_encoding is None
+                and self.scope["method"] != "HEAD"
+                and status_code not in (204, 304)
+            ):
                 # Neither content-length nor transfer-encoding specified
                 self.chunked_encoding = True
                 content.append(b"transfer-encoding: chunked\r\n")
@@ -492,7 +480,10 @@ class RequestResponseCycle:
             if self.scope["method"] == "HEAD":
                 self.expected_content_length = 0
             elif self.chunked_encoding:
-                content = [b"%x\r\n" % len(body), body, b"\r\n"]
+                if body:
+                    content = [b"%x\r\n" % len(body), body, b"\r\n"]
+                else:
+                    content = []
                 if not more_body:
                     content.append(b"0\r\n\r\n")
                 self.transport.write(b"".join(content))
@@ -520,12 +511,13 @@ class RequestResponseCycle:
 
     async def receive(self):
         if self.waiting_for_100_continue and not self.transport.is_closing():
-            self.transport.write(b"HTTP/1.1 100 Continue\r\n")
+            self.transport.write(b"HTTP/1.1 100 Continue\r\n\r\n")
             self.waiting_for_100_continue = False
 
-        self.flow.resume_reading()
-        await self.message_event.wait()
-        self.message_event.clear()
+        if not self.disconnected and not self.response_complete:
+            self.flow.resume_reading()
+            await self.message_event.wait()
+            self.message_event.clear()
 
         if self.disconnected or self.response_complete:
             message = {"type": "http.disconnect"}
